package com.sd.business.service.production.impl;

import cn.hutool.core.util.ObjectUtil;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.ruoyi.common.constant.StatusConstant;
import com.ruoyi.common.core.domain.BaseIdPo;
import com.ruoyi.common.utils.SecurityUtils;
import com.ruoyi.common.utils.wrapper.IWrapper;
import com.sd.business.entity.bom.bo.BomSpecBo;
import com.sd.business.entity.bom.po.BomSpec;
import com.sd.business.entity.department.constant.DepartmentConstant;
import com.sd.business.entity.department.po.Department;
import com.sd.business.entity.in.dto.InOutStorageDto;
import com.sd.business.entity.in.emums.InOutTypeEnum;
import com.sd.business.entity.in.emums.OutDetailTypeEnum;
import com.sd.business.entity.in.po.InOutStorageBom;
import com.sd.business.entity.order.enums.OrderExceptionTypeEnum;
import com.sd.business.entity.order.enums.OrderStatusEnum;
import com.sd.business.entity.order.po.OrderInfo;
import com.sd.business.entity.order.po.OrderSku;
import com.sd.business.entity.order.po.OrderSkuBom;
import com.sd.business.entity.production.dto.StockPreparationDto;
import com.sd.business.entity.production.po.ProductionTask;
import com.sd.business.entity.production.po.ProductionWorkOrder;
import com.sd.business.entity.production.vo.PackageBomVo;
import com.sd.business.entity.production.vo.StockPreparationVo;
import com.sd.business.entity.sku.po.Sku;
import com.sd.business.entity.sku.po.SkuSpec;
import com.sd.business.entity.warehouse.constant.WarehouseConstant;
import com.sd.business.mapper.production.StockPreparationMapper;
import com.sd.business.service.bom.BomSpecService;
import com.sd.business.service.in.InOutStorageService;
import com.sd.business.service.order.OrderService;
import com.sd.business.service.order.OrderSkuBomService;
import com.sd.business.service.order.OrderSkuService;
import com.sd.business.service.production.ProductionTaskService;
import com.sd.business.service.production.ProductionWorkOrderService;
import com.sd.business.service.production.StockPreparationService;
import com.sd.business.service.sku.SkuService;
import com.sd.business.service.sku.SkuSpecService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Service
public class StockPreparationServiceImpl implements StockPreparationService {

    @Resource
    private StockPreparationMapper stockPreparationMapper;

    @Autowired
    private OrderService orderService;

    @Autowired
    private OrderSkuService orderSkuService;

    @Autowired
    private OrderSkuBomService orderSkuBomService;

    @Autowired
    private ProductionTaskService productionTaskService;

    @Autowired
    private ProductionWorkOrderService productionWorkOrderService;

    @Autowired
    private SkuService skuService;

    @Autowired
    private SkuSpecService skuSpecService;

    @Autowired
    private InOutStorageService inOutStorageService;

    @Autowired
    private BomSpecService bomSpecService;

    @Override
    public List<StockPreparationVo> uncompletedList(StockPreparationDto dto) {
        IWrapper<StockPreparationVo> wrapper = createWrapper(dto);
        wrapper.eq("oi", OrderInfo::getStatus, OrderStatusEnum.STOCK_PREPARATION.getKey());
        wrapper.eq("os", OrderSku::getStockPreparationStatus, StatusConstant.NO);
        wrapper.eq("oi", OrderInfo::getExceptionType, OrderExceptionTypeEnum.NORMAL.getKey().toString());
        return stockPreparationMapper.uncompletedList(wrapper);
    }

    @Override
    public Page<StockPreparationVo> completedPage(StockPreparationDto dto) {
        IWrapper<StockPreparationVo> wrapper = createWrapper(dto);
        wrapper.eq("os", OrderSku::getStockPreparationStatus, StatusConstant.YES);
        wrapper.orderByDesc("os", OrderSku::getStockPreparationTime);
        wrapper.orderByAsc("os", OrderSku::getId);
        return stockPreparationMapper.completedPage(dto.getPage(), wrapper);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public synchronized void submit(List<Long> orderSkuIdList) {

        // 更爱状态为已备料
        List<OrderSku> orderSkuList = orderSkuService.list(q -> q
                .eq(OrderSku::getStockPreparationStatus, StatusConstant.NO)
                .in(BaseIdPo::getId, orderSkuIdList));

        if (orderSkuList.size() == 0) {
            return;
        }

        // 订单sku出库
        Date stockPreparationTime = new Date();
        orderSkuList.forEach(item -> {
            item.setStockPreparationTime(stockPreparationTime);
            item.setStockPreparationStatus(StatusConstant.YES);
        });
        orderSkuService.updateBatchById(orderSkuList);

        // 订单id列表
        List<Long> orderIdList = orderSkuList.stream().map(OrderSku::getOrderId).distinct().collect(Collectors.toList());

        // 订单出库
        orderDelivery(orderIdList, orderSkuList);

        List<OrderInfo> orderInfoList = new ArrayList<>();
        List<ProductionTask> productionTaskList = new ArrayList<>();
        List<ProductionWorkOrder> productionWorkOrderList = new ArrayList<>();

        // 生产任务、生产工单
        Map<Long, List<OrderSku>> map = orderSkuList.stream().collect(Collectors.groupingBy(OrderSku::getOrderId));
        map.forEach((orderId, itemOrderSkuList) -> {
            // 订单状态改为 生产中
            addOrderInfo(orderId, orderInfoList);
            // 添加 生产任务、工单
            addTaskAndWorkOrder(orderId, itemOrderSkuList, productionTaskList, productionWorkOrderList);
        });

        // 保存数据
        if (orderInfoList.size() > 0) {
            orderService.updateBatchById(orderInfoList);
        }
        if (productionTaskList.size() > 0) {
            productionTaskService.saveBatch(productionTaskList);
        }
        if (productionWorkOrderList.size() > 0) {
            productionWorkOrderService.saveBatch(productionWorkOrderList);
        }

    }

    @Override
    public List<PackageBomVo> getPackageBomList(List<Long> orderSkuIdList) {

        List<OrderSkuBom> list = orderSkuBomService.list(q -> q.in(OrderSkuBom::getOrderSkuId, orderSkuIdList));

        List<PackageBomVo> result = new ArrayList<>(list.stream()
                .map(item -> PackageBomVo.builder().bomSpecId(item.getBomSpecId()).quantity(item.getQuantity()).build())
                .collect(Collectors.toMap(
                        PackageBomVo::getBomSpecId,
                        Function.identity(),
                        (v1, v2) -> {
                            v1.setQuantity(v1.getQuantity().add(v2.getQuantity()));
                            return v1;
                        }
                )).values());

        bomSpecService.attributeAssign(result, PackageBomVo::getBomSpecId, (item, bomSpec) -> {
            item.setBomSpecCode(bomSpec.getCode());
            item.setBomSpecName(bomSpec.getName());
        });

        return result;
    }

    private IWrapper<StockPreparationVo> createWrapper(StockPreparationDto dto) {
        IWrapper<StockPreparationVo> wrapper = IWrapper.getWrapper();

        wrapper.eq("oi", OrderInfo::getCode, dto.getOrderCode());
        wrapper.eq("oi", OrderInfo::getWlnCode, dto.getOrderWlnCode());

        wrapper.eq("bs", BomSpec::getCode, dto.getBomSpecCode());
        wrapper.eq("bs", BomSpec::getName, dto.getBomSpecName());

        wrapper.eq("d", Department::getId, dto.getDepartmentId());

        wrapper.eq("ss", SkuSpec::getCode, dto.getSkuSpecCode());
        wrapper.eq("ss", SkuSpec::getName, dto.getSkuSpecName());

        wrapper.eq("bs", BomSpec::getWidth, dto.getWidth());

        return wrapper;
    }

    /**
     * 订单状态改为 生产中
     */
    private void addOrderInfo(Long orderId, List<OrderInfo> orderInfoList) {
        OrderInfo orderInfo = new OrderInfo();
        orderInfo.setId(orderId);
        orderInfo.setStatus(OrderStatusEnum.IN_PRODUCTION.getKey());
        orderInfoList.add(orderInfo);
    }

    /**
     * 添加 生产任务、工单
     */
    private void addTaskAndWorkOrder(Long orderId,
                                     List<OrderSku> orderSkuList,
                                     List<ProductionTask> productionTaskList,
                                     List<ProductionWorkOrder> productionWorkOrderList) {

        OrderInfo orderInfo = orderService.getById(orderId);
        Map<String, ProductionTask> skuSpecIdProductionTaskMap = new HashMap<>();
        int num = 1;

        for (OrderSku orderSku : orderSkuList) {

            Long skuSpecId = orderSku.getSkuSpecId();
            Long bomSpecId = orderSku.getBomSpecId();
            Integer printType = orderSku.getPrintType();
            BigDecimal quantity = orderSku.getQuantity();

            ProductionTask productionTask = skuSpecIdProductionTaskMap.computeIfAbsent(
                    skuSpecId + ":" + bomSpecId + ":" + printType,
                    item -> {
                        SkuSpec skuSpec = skuSpecService.getById(skuSpecId);
                        Sku sku = skuService.getById(skuSpec.getSkuId());

                        ProductionTask tempProductionTask = new ProductionTask();
                        tempProductionTask.setId(IdWorker.getId());
                        tempProductionTask.setOrderId(orderSku.getOrderId());
                        tempProductionTask.setSkuSpecId(skuSpecId);
                        tempProductionTask.setBomSpecId(bomSpecId);
                        tempProductionTask.setCraftProductionLineId(sku.getCraftProductionLineId());
                        tempProductionTask.setMachinedPanel(skuSpec.getMachinedPanel());
                        tempProductionTask.setPrintType(printType);
                        tempProductionTask.setProductionQuantity(BigDecimal.ZERO);
                        tempProductionTask.setCompleteQuantity(BigDecimal.ZERO);

                        productionTaskList.add(tempProductionTask);

                        return tempProductionTask;
                    });

            productionTask.setProductionQuantity(productionTask.getProductionQuantity().add(quantity));

            for (int i = 0; i < quantity.intValue(); i++) {
                ProductionWorkOrder productionWorkOrder = new ProductionWorkOrder();
                productionWorkOrder.setProductionTaskId(productionTask.getId());
                productionWorkOrder.setCode(orderInfo.getCode() + String.format("-%05d", num));
                productionWorkOrder.setStatus(0);
                productionWorkOrderList.add(productionWorkOrder);
                num++;
            }

        }
    }

    /**
     * 订单出库
     */
    private void orderDelivery(List<Long> orderIdList, List<OrderSku> orderSkuList) {

        List<OrderSkuBom> orderSkuBomList = orderSkuBomService.list(q -> q.in(OrderSkuBom::getOrderId, orderIdList));

        // 合并sku主材和包材
        List<InOutStorageBom> list = new ArrayList<>(
                Stream.concat(
                                orderSkuList.stream().map(item -> {
                                    InOutStorageBom inOutStorageBom = new InOutStorageBom();
                                    inOutStorageBom.setBomSpecId(item.getBomSpecId());
                                    inOutStorageBom.setQuantity(item.getQuantity());
                                    return inOutStorageBom;
                                }),
                                orderSkuBomList.stream().map(item -> {
                                    InOutStorageBom inOutStorageBom = new InOutStorageBom();
                                    inOutStorageBom.setBomSpecId(item.getBomSpecId());
                                    inOutStorageBom.setQuantity(item.getQuantity());
                                    return inOutStorageBom;
                                })
                        )
                        .collect(Collectors.toMap(
                                        InOutStorageBom::getBomSpecId,
                                        Function.identity(),
                                        (v1, v2) -> {
                                            v1.setQuantity(v1.getQuantity().add(v2.getQuantity()));
                                            return v1;
                                        }
                                )
                        ).values());

        List<Long> bomSpecIdList = list.stream().map(InOutStorageBom::getBomSpecId).collect(Collectors.toList());
        Map<Long, BomSpecBo> bomSpecBo = skuSpecService.getBomSpecBoByIdList(bomSpecIdList);

        // 主材从半成品仓出库
        InOutStorageDto semiFinishedProduct = new InOutStorageDto();
        semiFinishedProduct.setType(InOutTypeEnum.OUT.getKey());
        semiFinishedProduct.setDetailType(OutDetailTypeEnum.PRODUCTION.getKey());
        semiFinishedProduct.setWarehouseId(WarehouseConstant.SEMI_FINISHED_PRODUCT);
        semiFinishedProduct.setDepartmentId(DepartmentConstant.SD_SPORTS);
        semiFinishedProduct.setApplicant(SecurityUtils.getLoginUser().getUser().getNickName());
        List<InOutStorageBom> semiFinishedProductInOutStorageBomList = list.stream().filter(item -> {
            BomSpecBo tempBomSpecBo = bomSpecBo.get(item.getBomSpecId());
            if (tempBomSpecBo == null) {
                return false;
            }
            return ObjectUtil.equal(tempBomSpecBo.getClassifyParentId(), 1L);
        }).collect(Collectors.toList());
        semiFinishedProduct.setInOutStorageBomList(semiFinishedProductInOutStorageBomList);
        inOutStorageService.add(semiFinishedProduct);

        // 包材从包材仓出库
        InOutStorageDto packagingMaterial = new InOutStorageDto();
        packagingMaterial.setType(InOutTypeEnum.OUT.getKey());
        packagingMaterial.setDetailType(OutDetailTypeEnum.PRODUCTION.getKey());
        packagingMaterial.setWarehouseId(WarehouseConstant.PACKAGING_MATERIAL);
        packagingMaterial.setDepartmentId(DepartmentConstant.SD_SPORTS);
        packagingMaterial.setApplicant(SecurityUtils.getLoginUser().getUser().getNickName());
        List<InOutStorageBom> packagingMaterialInOutStorageBomList = list.stream().filter(item -> {
            BomSpecBo tempBomSpecBo = bomSpecBo.get(item.getBomSpecId());
            if (tempBomSpecBo == null) {
                return true;
            }
            return ObjectUtil.notEqual(tempBomSpecBo.getClassifyParentId(), 1L);
        }).collect(Collectors.toList());
        packagingMaterial.setInOutStorageBomList(packagingMaterialInOutStorageBomList);
        inOutStorageService.add(packagingMaterial);

    }

}
